from bxcommon.feed.raw_block import AbstractRawBlock
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo


class EthRawBlock(AbstractRawBlock[InternalEthBlockInfo]):
    """
    EthBlock Notification Object

    """
    pass
