class GatewayStartArgs:
    GATEWAY_RPC_PORT = "gateway_rpc_port"
    GATEWAY_RPC_HOST = "gateway_rpc_host"
    GATEWAY_RPC_USER = "gateway_rpc_user"
    GATEWAY_RPC_PASSWORD = "gateway_rpc_password"
    DEFAULT_TX_FLAG = "default_tx_flag"
    HTTP_INFO_PORT = "http_info_port"
    CONSENSUS_PORT = "consensus_port"
    RELAY_STATE = "relay_state"
    IS_CONSENSUS = "is_consensus"


