class OntMessageType:
    VERSION = b"version"
    VERACK = b"verack"
    GET_ADDRESS = b"getaddr"
    ADDRESS = b"addr"
    PING = b"ping"
    PONG = b"pong"
    GET_HEADERS = b"getheaders"
    HEADERS = b"headers"
    INVENTORY = b"inv"
    GET_DATA = b"getdata"
    BLOCK = b"block"
    TRANSACTIONS = b"tx"
    CONSENSUS = b"consensus"
    GET_BLOCKS = b"getblocks"
    NOT_FOUND = b"notfound"
    DISCONNECT = b"disconnect"
