class NodeStartArgs:
    EXTERNAL_PORT = "external_port"
    NON_SSL_PORT = "non_ssl_port"
    SDN_ROOT_URL = "sdn_root_url"
    CA_CERT_URL = "ca_cert_url"
    PRIVATE_SSL_BASE_URL = "private_ssl_base_url"
