BTC_SHA_HASH_LEN = 32

TX_VERSION_LEN = 4
TX_SEGWIT_FLAG_LEN = 2
TX_LOCK_TIME_LEN = 4
TX_SEGWIT_FLAG_VALUE = 1

BTC_VARINT_SHORT_INDICATOR = 0xFD
BTC_VARINT_INT_INDICATOR = 0xFE
BTC_VARINT_LONG_INDICATOR = 0xFF
