from cryptography.x509 import ObjectIdentifier


class ExtensionsObjectIds:
    NODE_TYPE = ObjectIdentifier("1.22.333.4444")
    NODE_ID = ObjectIdentifier("1.22.333.4445")
    ACCOUNT_ID = ObjectIdentifier("1.22.333.4446")
    NODE_PRIVILEGES = ObjectIdentifier("1.22.333.4447")
