ONT_HASH_LEN = 32

ONT_TX_DEPLOY_TYPE_INDICATOR = 0xD0
ONT_TX_DEPLOY_TYPE_INDICATOR_AS_BYTEARRAY = bytearray([ONT_TX_DEPLOY_TYPE_INDICATOR])
ONT_TX_INVOKE_TYPE_INDICATOR = 0xD1
ONT_TX_INVOKE_TYPE_INDICATOR_AS_BYTEARRAY = bytearray([ONT_TX_INVOKE_TYPE_INDICATOR])
