GENERAL_CATEGORY = "General category"
PROCESSING_FAILED_CATEGORY = "Processing problem"
LOGGING_CATEGORY = "Logging problem"
MEMORY_CATEGORY = "Memory problem"
REQUEST_RESPONSE_CATEGORY = "Unsatisfactory data in the response"
NETWORK_CATEGORY = "Networking problem"
CONNECTION_PROBLEM_CATEGORY = "Connection problem"
THREAD_CATEGORY = "Threading problems"
UNCATEGORIZED = "Uncategorized"
THIRD_PARTY_CATEGORY = "Third party log message"
FAILED_TO_ACQUIRE_REDIS_LOCK = "Failed to acquire redis lock"
NETWORK_ERROR = "Network error"
NODE_ID_ERROR = "Missing or incorrect node id"
MISSING_NODE = "Missing node"
EXCEPTION_OR_CANCELLATION = "Exception error or task cancelled"
EVENT_ERROR = "Event missing, incorrect or incompatible"
CONFIGURATION_ERROR = "Error setting up or reading configuration"
GENERAL_WARNING = "General warning category"
AUTHENTICATION_ERROR = "Authentication error"
RPC_ERROR = "Rpc error"
AUDITING_ERROR = "Auditing error"
NOTIFICATION_FROM_RELAY_CATEGORY = "Notification from relay"
